#' ---
#' title: "Biases in mediation"
#' author: ""
#' date: "`r Sys.time()`"
#' output:
#'  html_document:
#'    toc: true
#'    number_sections: true
#'    toc_float: true
#'    theme: readable
#'    highlight: tango
#'    df_print: paged
#'    code_folding: show
#' ---
#'
#' These analyses demonstrate with simple examples the correctness
#' of the bias formulas given in 
#' [Savalai (2019)](https://doi.org/10.1037/met0000181). Based on 
#' true score effects for a simple mediation model and reliability 
#' estimates, a SEM is fitted to the attenuated implied variance-
#' covariance matrix.The difference in the estimated effects to the 
#' data-generating true effects give the same results as the analytic 
#' bias formulas.
#' 


# Clear workspace
rm(list = ls())

# Load packages
suppressPackageStartupMessages({
  library("lavaan")
})



#
# Analytic bias for regressions weights
#
# Note: Assumes variances of 1 for the true scores.
#
# @param rxx       A reliability estimate for X
# @param rmm       A reliability estimate for M
# @param MONX      A standardized regression weight of M on X
# @param YONM      A standardized regression weight of Y on M
# @param YONX      A standardized regression weight of Y on X
# @return          A numeric vector
# @source          https://doi.org/10.1037/met0000181
#
bias_a <- function(rxx = 1, rmm = 1, ryy = 1, 
                   MONX, YONM, YONX) {

  # eq A6
  a <- rxx * MONX
  
  # eq A7a
  b <- rmm * (YONM + MONX * YONX - rxx * (MONX * YONM + YONX) * MONX) /
    (1 - rxx * rmm * MONX^2)

  # eq A8
  c <- rxx * ((YONX + MONX * YONM) - MONX * rmm * (YONM + MONX * YONX)) /
    (1 - rxx * rmm * MONX^2)
  
  # biases
  out <- c(bias_a = MONX - a, bias_b = YONM - b, bias_c = YONX - c)
    
  return(out)
  
}


#
# Generate attenuated variance-covariance matrix from 
#   standardized regression weights of a mediation model
#
# @param a     A standardized regression weight of M on X
# @param b     A standardized regression weight of Y on X
# @param c     A standardized regression weight of Y on M
# @param rxx   Reliability of X
# @param rmm   Reliability of M
# @param ryy   Reliability of Y
# @return      A 3x3 variance-covariance matrix
#
get_r <- function(a, b, c,
                  rxx = 1, rmm = 1, ryy = 1) {
  
  m <- diag(1 / c(rxx, rmm, ryy))
  m[lower.tri(m)] <- m[upper.tri(m)] <- c(a, a * b + c, b + a * c)
  colnames(m) <- rownames(m) <- c("X", "M", "Y")
  return(m)
  
}


#'
#' # rxx = 1, rmm = 1, ryy = .70
#'
#' Unreliability in Y does not lead to bias.
#'

# Attenuated effects
(b <- coef(sem(model = "M ~ X; Y ~ M + X;", 
               sample.cov = get_r(a = .3, b = .3, c = .3, 
                                  rxx = 1, rmm = 1, ryy = .70), 
               sample.nobs = 1000)))

# Analytic bias
bias_a(ryy = .70, MONX = .30, YONM = .30, YONX = .30)

# True effect minus attenuated effect corresponds to analytic bias
.30 - b["M~X"]
.30 - b["Y~M"]
.30 - b["Y~X"]


#'
#' # rxx = 1, rmm = .70, ryy = 1
#' 
#' Unreliability in M does not bias a, but leads to biases in b and c.
#'

# Attenuated effects
(b <- coef(sem(model = "M ~ X; Y ~ M + X;", 
               sample.cov = get_r(a = .3, b = .3, c = .3, 
                                  rxx = 1, rmm = .70, ryy = 1), 
               sample.nobs = 1000)))

# Analytic bias
bias_a(rmm = .70, MONX = .30, YONM = .30, YONX = .30)

# True effect minus attenuated effect corresponds to analytic bias
.30 - b["M~X"]
.30 - b["Y~M"]
.30 - b["Y~X"]


#'
#' # rxx = .70, rmm = 1, ryy = 1
#' 
#' Unreliability in X leads to biases in all effects.
#'

# Attenuated effects
(b <- coef(sem(model = "M ~ X; Y ~ M + X;", 
               sample.cov = get_r(a = .3, b = .3, c = .3, 
                                  rxx = .70, rmm = 1, ryy = 1), 
               sample.nobs = 1000)))

# Analytic bias
bias_a(rxx = .70, MONX = .30, YONM = .30, YONX = .30)

# True effect minus attenuated effect corresponds to analytic bias
.30 - b["M~X"]
.30 - b["Y~M"]
.30 - b["Y~X"]


