#' ---
#' title: "Parameter distributions"
#' author: ""
#' date: "`r Sys.time()`"
#' output:
#'    html_document:
#'        toc: true
#'        number_sections: true
#'        toc_float: true
#'        theme: readable
#'        highlight: tango
#'        df_print: paged
#'        code_folding: show
#' ---
#'
#' The simulation relied on several distributions to 
#' randomly draw values for various settings such as
#' the sample size or random effects. These distributions 
#' are visualized here for demonstration.
#' 


# Clear workspace
rm(list = ls())

suppressPackageStartupMessages({
  library("here")
  library("openxlsx")
})



#'
#' # Random effect (SD)
#'
#' [van Erp et al. (2017)](http://doi.org/10.5334/jopd.33) 
#' collected between-study heterogeneity estimates of 
#' meta-analyses published in *Psychological Bulletin* between 
#' 1990 and 2013. The respective [database](https://osf.io/wyhve/) 
#' was provided under a public domain license. Here, the 
#' distributions of these estimates are summarized for 
#' meta-analyses using correlations as effect sizes.
#' 

# Read database
dat <- openxlsx::read.xlsx(
  here::here("data", "Data 1990-2013 with tau values.xlsx"),
  sheet = 1
)

# Select entries with correlations
dat <- dat[!is.na(dat$'tau'), ]
dat <- dat[dat$'Type.of.ES' %in% c("Weighted Pearson's r", 
                                   "Pearson's r", 
                                   "Unweighted Pearson's r"), ]

# Number of estimates
length(dat$tau)

# Distribution of estimates
hist(dat$tau, main = "Random effects", xlab = "tau", prob = TRUE)

# Distribution used for simulation
tau <- abs(rnorm(100000, 0, .20))
repeat {
  tau[tau > .50] <- tau[tau > .50] / 2
  if (all(tau <= .50)) break
}
lines(density(tau))

# Quantiles of estimates
round(quantile(dat$tau, c(.05, .25, .50, .75, .95)), 2)
round(quantile(tau, c(.05, .25, .50, .75, .95)), 2)
range(dat$tau)
range(tau)
rm(tau)



#'
#' # Sample size
#'

# Random draws
n <- rgamma(100000, 1.3, 1 / 250) + 50
repeat {
  n[n > 900] <- n[n > 900] / 2
  if (all(n <= 900)) break
}
n <- round(n)

# Distribution of estimates
hist(n, main = "Sample size", xlab = "Sample size", prob = TRUE)

# Quantiles of estimates
round(quantile(n, c(.05, .25, .50, .75, .95)), 2)
range(n)
rm(n)



#'
#' # Number of effects in meta-analysis
#'

# Random draws
k <- round(runif(100000, 5, 50))

# Distribution of estimates
hist(k, main = "Number of effects", xlab = "Number of effects", prob = TRUE)

# Quantiles of estimates
round(quantile(k, c(.05, .25, .50, .75, .95)), 2)
range(k)
rm(k)



#'
#' # Reliabilities of measures
#'

# Random draws
rxx <- round(runif(100000, .65, .95), 2)

# Distribution of estimates
hist(rxx, main = "Reliability", xlab = "Reliability", prob = TRUE)

# Quantiles of estimates
round(quantile(rxx, c(.05, .25, .50, .75, .95)), 2)
range(rxx)
rm(rxx)


